/**
 * The atomic module is intended to provide some basic support for lock-free
 * concurrent programming.  Some common operations are defined, each of which
 * may be performed using the specified memory barrier or a less granular
 * barrier if the hardware does not support the version requested.  This
 * model is based on a design by Alexander Terekhov as outlined in
 * <a href=http://groups.google.com/groups?threadm=3E4820EE.6F408B25%40web.de>
 * this thread</a>.
 *
 * Design Issues:
 *
 * Originally, all functions were intended to be either static or non-static
 * members of the Atomic class.  However, DMD currently can not disambiguate
 * between the two if they have the same name, so the current design was
 * chosen.  This design also seems to be a bit more readable, so it may remain
 * even if the compiler bug (?) is fixed.
 *
 * Future Directions:
 *
 * As many architectures provide optimized atomic support for common integer
 * operations (such as increment, decrement, etc), these may be added in some
 * form--perhaps as additional global functions and in an AtomicValue type.
 */
module std.atomic;


////////////////////////////////////////////////////////////////////////////////
// Synchronization Options
////////////////////////////////////////////////////////////////////////////////


/**
 * Memory synchronization flag.  If the supplied option is not available on the
 * current platform then a stronger method will be used instead.
 */
enum msync
{
    hlb,    /// hoist-load barrier
    ddhlb,  /// hoist-load barrier with data-dependency "hint"
    hsb,    /// hoist-store barrier
    slb,    /// sink-load barrier
    ssb,    /// sink-store barrier
    acq,    /// hoist-load + hoist-store barrier
    rel,    /// sink-load + sink-store barrier
    none    /// naked
}


////////////////////////////////////////////////////////////////////////////////
// Internal Type Checking
////////////////////////////////////////////////////////////////////////////////


private
{
    template isIntegerType( T )
    {
        const bool isIntegerType = is( T == byte )  || is( T == ubyte )  ||
                                   is( T == short ) || is( T == ushort ) ||
                                   is( T == int )   || is( T == uint )   ||
                                   is( T == long )  || is( T == ulong );
    }


    template isPointerType( T )
    {
        const bool isPointerType = false;
    }


    template isPointerType( T : T* )
    {
        const bool isPointerType = true;
    }


    template isValidAtomicType( T )
    {
        const bool isValidAtomicType = true;
    }


    template isValidNumericType( T )
    {
        const bool isValidNumericType = isIntegerType!( T ) ||
                                        isPointerType!( T );
    }


    template suppliedSizeofMatchesTrueSizeof( T, size_t sizeofT )
    {
        const bool suppliedSizeofMatchesTrueSizeof = sizeofT == T.sizeof;
    }


    template isHoistOp( msync ms )
    {
        const bool isHoistOp = ms == msync.hlb   ||
                               ms == msync.ddhlb ||
                               ms == msync.hsb   ||
                               ms == msync.acq;
    }


    template isSinkOp( msync ms )
    {
        const bool isSinkOp = ms == msync.slb ||
                              ms == msync.ssb ||
                              ms == msync.rel;
    }
}


////////////////////////////////////////////////////////////////////////////////
// x86 Atomic Function Implementation
////////////////////////////////////////////////////////////////////////////////


private
{
    version( X86 )
    {
        ////////////////////////////////////////////////////////////////////////
        // x86 Value Requirements
        ////////////////////////////////////////////////////////////////////////


        //
        // NOTE: Strictly speaking, the x86 supports atomic operations on
        //       unaligned values.  However, this is far slower than the
        //       common case, so such behavior should be prohibited.
        //
        template atomicValueIsProperlyAligned( T )
        {
            bool atomicValueIsProperlyAligned( size_t addr )
            {
                return addr % T.sizeof == 0;
            }
        }


        ////////////////////////////////////////////////////////////////////////
        // Atomic Load
        ////////////////////////////////////////////////////////////////////////


        template doAtomicLoad( T, bool membar, size_t sizeofT : 1 = T.sizeof )
        {
            T doAtomicLoad( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, val;
                        lock;
                        mov AL, [EAX];
                    }
                }
                else
                {
                    volatile
                    {
                        return val;
                    }
                }
            }
        }


        template doAtomicLoad( T, bool membar, size_t sizeofT : 2 = T.sizeof )
        {
            T doAtomicLoad( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, val;
                        lock;
                        mov AX, [EAX];
                    }
                }
                else
                {
                    volatile
                    {
                        return val;
                    }
                }
            }
        }


        template doAtomicLoad( T, bool membar, size_t sizeofT : 4 = T.sizeof )
        {
            T doAtomicLoad( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, val;
                        lock;
                        mov EAX, [EAX];
                    }
                }
                else
                {
                    volatile
                    {
                        return val;
                    }
                }
            }
        }


        version( X86_64 )
        {
            template doAtomicLoad( T, bool membar, size_t sizeofT : 8 = T.sizeof )
            {
                T doAtomicLoad( inout T val )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                    static if( membar )
                    {
                        volatile asm
                        {
                            mov RAX, val;
                            lock;
                            mov RAX, [RAX];
                        }
                    }
                    else
                    {
                        volatile
                        {
                            return val;
                        }
                    }
                }
            }
        }
        else
        {
            template doAtomicLoad( T, bool membar, size_t sizeofT : 8 = T.sizeof )
            {
                T doAtomicLoad( inout T val )
                {
                    const bool operationOnlyAvailableFor64bit = false;
                    static assert( operationOnlyAvailableFor64bit );
                }
            }
        }


        ////////////////////////////////////////////////////////////////////////
        // Atomic Store
        ////////////////////////////////////////////////////////////////////////


        template doAtomicStore( T, bool membar, size_t sizeofT : 1 = T.sizeof )
        {
            void doAtomicStore( inout T val, T newval )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        mov BL, newval;
                        lock;
                        xchg [EAX], BL;
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        mov BL, newval;
                        mov [EAX], BL;
                    }
                }
            }
        }


        template doAtomicStore( T, bool membar, size_t sizeofT : 2 = T.sizeof )
        {
            void doAtomicStore( inout T val, T newval )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        mov BX, newval;
                        lock;
                        xchg [EAX], BX;
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        mov BX, newval;
                        mov [EAX], BX;
                    }
                }
            }
        }


        template doAtomicStore( T, bool membar, size_t sizeofT : 4 = T.sizeof )
        {
            void doAtomicStore( inout T val, T newval )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, val;
                        mov EBX, newval;
                        lock;
                        xchg [EAX], EBX;
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov EAX, val;
                        mov EBX, newval;
                        mov [EAX], EBX;
                    }
                }
            }
        }


        version( X86_64 )
        {
            template doAtomicStore( T, bool membar, size_t sizeofT : 8 = T.sizeof )
            {
                void doAtomicStore( inout T val, T newval )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                    static if( membar )
                    {
                        volatile asm
                        {
                            mov RAX, val;
                            mov RBX, newval;
                            lock;
                            xchg [RAX], RBX;
                        }
                    }
                    else
                    {
                        volatile asm
                        {
                            mov RAX, val;
                            mov RBX, newval;
                            mov [RAX], RBX;
                        }
                    }
                }
            }
        }
        else
        {
            template doAtomicStore( T, bool membar, size_t sizeofT : 8 = T.sizeof )
            {
                void doAtomicStore( inout T val, T newval )
                {
                    const bool operationOnlyAvailableFor64bit = false;
                    static assert( operationOnlyAvailableFor64bit );
                }
            }
        }


        ////////////////////////////////////////////////////////////////////////
        // Atomic Store If
        ////////////////////////////////////////////////////////////////////////


        template doAtomicStoreIf( T, bool membar, size_t sizeofT : 1 = T.sizeof )
        {
            bool doAtomicStoreIf( inout T val, T newval, T equalTo )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    volatile asm
                    {
                        mov BL, newval;
                        mov AL, equalTo;
                        mov ECX, pval;
                        lock;
                        cmpxchg [ECX], BL;
                        setz AL;
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov BL, newval;
                        mov AL, equalTo;
                        mov ECX, pval;
                        lock; // lock needed to make this op atomic
                        cmpxchg [ECX], BL;
                        setz AL;
                    }
                }
            }
        }


        template doAtomicStoreIf( T, bool membar, size_t sizeofT : 2 = T.sizeof )
        {
            bool doAtomicStoreIf( inout T val, T newval, T equalTo )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    volatile asm
                    {
                        mov BX, newval;
                        mov AX, equalTo;
                        mov ECX, pval;
                        lock;
                        cmpxchg [ECX], BX;
                        setz AL;
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov BX, newval;
                        mov AX, equalTo;
                        mov ECX, pval;
                        lock; // lock needed to make this op atomic
                        cmpxchg [ECX], BX;
                        setz AL;
                    }
                }
            }
        }


        template doAtomicStoreIf( T, bool membar, size_t sizeofT : 4 = T.sizeof )
        {
            bool doAtomicStoreIf( inout T val, T newval, T equalTo )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    volatile asm
                    {
                        mov EBX, newval;
                        mov EAX, equalTo;
                        mov ECX, val;
                        lock;
                        cmpxchg [ECX], EBX;
                        setz AL;
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov EBX, newval;
                        mov EAX, equalTo;
                        mov ECX, val;
                        lock; // lock needed to make this op atomic
                        cmpxchg [ECX], EBX;
                        setz AL;
                    }
                }
            }
        }


        version( X86_64 )
        {
            template doAtomicStoreIf( T, bool membar, size_t sizeofT : 8 = T.sizeof )
            {
                bool doAtomicStoreIf( inout T val, T newval, T equalTo )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                    static if( membar )
                    {
                        volatile asm
                        {
                            mov RBX, newval;
                            mov RAX, equalTo;
                            mov RCX, val;
                            lock;
                            cmpxchg [RCX], RBX;
                            setz AL;
                        }
                    }
                    else
                    {
                        volatile asm
                        {
                            mov RBX, newval;
                            mov RAX, equalTo;
                            mov RCX, val;
                            lock; // lock needed to make this op atomic
                            cmpxchg [RCX], RBX;
                            setz AL;
                        }
                    }
                }
            }
        }
        else
        {
            template doAtomicStoreIf( T, bool membar, size_t sizeofT : 8 = T.sizeof )
            {
                bool doAtomicStoreIf( inout T val, T newval, T equalTo )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                    static if( membar )
                    {
                        volatile asm
                        {
                            lea	EDI, newval;
                            mov	EBX, [EDI];
                            mov	ECX, 4[EDI];
                            lea	EDI, equalTo;
                            mov	EAX, [EDI];
                            mov	EDX, 4[EDI];
                            mov	EDI, val;
                            lock;
                            cmpxch8b [EDI];
                            setz AL;
                        }
                    }
                    else
                    {
                        volatile asm
                        {
                            lea	EDI, newval;
                            mov	EBX, [EDI];
                            mov	ECX, 4[EDI];
                            lea	EDI, equalTo;
                            mov	EAX, [EDI];
                            mov	EDX, 4[EDI];
                            mov	EDI, val;
                            lock; // lock needed to make this op atomic
                            cmpxch8b [EDI];
                            setz AL;
                        }
                    }
                }
            }
        }


        ////////////////////////////////////////////////////////////////////////
        // Atomic Increment
        ////////////////////////////////////////////////////////////////////////


        template doAtomicIncrement( T, bool membar, size_t sizeofT : 1 = T.sizeof )
        {
            T doAtomicIncrement( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( isValidNumericType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        lock;
                        inc [EAX];
                        mov AL, [EAX];
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        lock; // lock needed to make this op atomic
                        inc [EAX];
                        mov AL, [EAX];
                    }
                }
            }
        }


        template doAtomicIncrement( T, bool membar, size_t sizeofT : 2 = T.sizeof )
        {

            T doAtomicIncrement( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( isValidNumericType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        lock;
                        inc [EAX];
                        mov AX, [EAX];
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        lock; // lock needed to make this op atomic
                        inc [EAX];
                        mov AX, [EAX];
                    }
                }
            }
        }


        template doAtomicIncrement( T, bool membar, size_t sizeofT : 4 = T.sizeof )
        {
            T doAtomicIncrement( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( isValidNumericType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, val;
                        lock;
                        inc [EAX];
                        mov EAX, [EAX];
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov EAX, val;
                        lock; // lock needed to make this op atomic
                        inc [EAX];
                        mov EAX, [EAX];
                    }
                }
            }
        }


        version( X86_64 )
        {
            template doAtomicIncrement( T, bool membar, size_t sizeofT : 8 = T.sizeof )
            {
                T doAtomicIncrement( inout T val )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T) );
                    static assert( isValidNumericType!(T) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                    static if( membar )
                    {
                        volatile asm
                        {
                            mov RAX, val;
                            lock;
                            inc [RAX];
                            mov RAX, [RAX];
                        }
                    }
                    else
                    {
                        volatile asm
                        {
                            mov RAX, val;
                            lock; // lock needed to make this op atomic
                            inc [RAX];
                            mov RAX, [RAX];
                        }
                    }
                }
            }
        }
        else
        {
            template doAtomicIncrement( T, bool membar, size_t sizeofT : 8 = T.sizeof )
            {
                void doAtomicIncrement( inout T val )
                {
                    const bool operationOnlyAvailableFor64bit = false;
                    static assert( operationOnlyAvailableFor64bit );
                }
            }
        }


        ////////////////////////////////////////////////////////////////////////
        // Atomic Decrement
        ////////////////////////////////////////////////////////////////////////


        template doAtomicDecrement( T, bool membar, size_t sizeofT : 1 = T.sizeof )
        {
            T doAtomicDecrement( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( isValidNumericType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        lock;
                        dec [EAX];
                        mov AL, [EAX];
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        lock; // lock needed to make this op atomic
                        dec [EAX];
                        mov AL, [EAX];
                    }
                }
            }
        }


        template doAtomicDecrement( T, bool membar, size_t sizeofT : 2 = T.sizeof )
        {
            T doAtomicDecrement( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( isValidNumericType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        lock;
                        dec [EAX];
                        mov AX, [EAX];
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov EAX, pval;
                        lock; // lock needed to make this op atomic
                        dec [EAX];
                        mov AX, [EAX];
                    }
                }
            }
        }


        template doAtomicDecrement( T, bool membar, size_t sizeofT : 4 = T.sizeof )
        {
            T doAtomicDecrement( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( isValidNumericType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    volatile asm
                    {
                        mov EAX, val;
                        lock;
                        dec [EAX];
                        mov EAX, [EAX];
                    }
                }
                else
                {
                    volatile asm
                    {
                        mov EAX, val;
                        lock; // lock needed to make this op atomic
                        dec [EAX];
                        mov EAX, [EAX];
                    }
                }
            }
        }


        version( X86_64 )
        {
            template doAtomicDecrement( T, bool membar, size_t sizeofT : 8 = T.sizeof )
            {
                T doAtomicDecrement( inout T val )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T) );
                    static assert( isValidNumericType!(T) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                    static if( membar )
                    {
                        volatile asm
                        {
                            mov RAX, val;
                            lock;
                            dec [RAX];
                            mov RAX, [RAX];
                        }
                    }
                    else
                    {
                        volatile asm
                        {
                            mov RAX, val;
                            lock; // lock needed to make this op atomic
                            dec [RAX];
                            mov RAX, [RAX];
                        }
                    }
                }
            }
        }
        else
        {
            template doAtomicDecrement( T, bool membar, size_t sizeofT : 8 = T.sizeof )
            {
                T doAtomicDecrement( inout T val )
                {
                    const bool operationOnlyAvailableFor64bit = false;
                    static assert( operationOnlyAvailableFor64bit );
                }
            }
        }
    }
}


////////////////////////////////////////////////////////////////////////////////
// x86 Atomic Function Accessors
////////////////////////////////////////////////////////////////////////////////


version( X86 )
{
    //
    // NOTE: x86 loads implicitly have acquire semantics so a membar is only necessary on release
    //

    /**
     * Refreshes the contents of 'val' from main memory.  This operation is both lock-free and atomic.
     *
     * Returns:
     *  The loaded value.
     *
     * Params:
     *  val = The value to load.  This value must be properly aligned.
     */
    template atomicLoad( T, msync ms : msync.none )  { alias doAtomicLoad!(T,isSinkOp!(ms)) atomicLoad; }
    template atomicLoad( T, msync ms : msync.hlb )   { alias doAtomicLoad!(T,isSinkOp!(ms)) atomicLoad; } /// ditto
    template atomicLoad( T, msync ms : msync.ddhlb ) { alias doAtomicLoad!(T,isSinkOp!(ms)) atomicLoad; } /// ditto
    template atomicLoad( T, msync ms : msync.acq )   { alias doAtomicLoad!(T,isSinkOp!(ms)) atomicLoad; } /// ditto

    //
    // NOTE: x86 stores implicitly have release semantics so a membar is only necessary on acquires
    //

    /**
     * Stores 'newval' to the memory referenced by 'val'.  This operation is both lock-free and atomic.
     *
     * Params:
     *  val     = The destination variable.
     *  newval  = The value to store.
     */
    template atomicStore( T, msync ms : msync.none ) { alias doAtomicStore!(T,isHoistOp!(ms)) atomicStore; }
    template atomicStore( T, msync ms : msync.ssb )  { alias doAtomicStore!(T,isHoistOp!(ms)) atomicStore; } /// ditto
    template atomicStore( T, msync ms : msync.acq )  { alias doAtomicStore!(T,isHoistOp!(ms)) atomicStore; } /// ditto
    template atomicStore( T, msync ms : msync.rel )  { alias doAtomicStore!(T,isHoistOp!(ms)) atomicStore; } /// ditto

    /**
     * Stores 'newval' to the memory referenced by 'val' if val is equal to 'equalTo'.  This operation
     * is both lock-free and atomic.
     *
     * Returns:
     *  true if the store occurred, false if not.
     *
     * Params:
     *  val     = The destination variable.
     *  newval  = The value to store.
     *  equalTo = The comparison value.
     */
    template atomicStoreIf( T, msync ms : msync.none ) { alias doAtomicStoreIf!(T,ms!=msync.none) atomicStoreIf; }
    template atomicStoreIf( T, msync ms : msync.ssb )  { alias doAtomicStoreIf!(T,ms!=msync.none) atomicStoreIf; } /// ditto
    template atomicStoreIf( T, msync ms : msync.acq )  { alias doAtomicStoreIf!(T,ms!=msync.none) atomicStoreIf; } /// ditto
    template atomicStoreIf( T, msync ms : msync.rel )  { alias doAtomicStoreIf!(T,ms!=msync.none) atomicStoreIf; } /// ditto

    /**
     * This operation is only legal for built-in value and pointer types, and is equivalent to an atomic
     * "val = val + 1" operation.  This function exists to facilitate use of the optimized increment
     * instructions provided by some architecures.  If no such instruction exists on the target platform
     * then the behavior will perform the operation using more traditional means.  This operation is both
     * lock-free and atomic.
     *
     * Returns:
     *  The result of an atomicLoad of val immediately following the increment operation.  This value
     *  is not required to be equal to the newly stored value.  Thus, competing writes are allowed to
     *  occur between the increment and successive load operation.
     *
     * Params:
     *  val = The value to increment.
     */
    template atomicIncrement( T, msync ms : msync.none ) { alias doAtomicIncrement!(T,ms!=msync.none) atomicIncrement; }
    template atomicIncrement( T, msync ms : msync.ssb )  { alias doAtomicIncrement!(T,ms!=msync.none) atomicIncrement; } /// ditto
    template atomicIncrement( T, msync ms : msync.acq )  { alias doAtomicIncrement!(T,ms!=msync.none) atomicIncrement; } /// ditto
    template atomicIncrement( T, msync ms : msync.rel )  { alias doAtomicIncrement!(T,ms!=msync.none) atomicIncrement; } /// ditto

    /**
     * This operation is only legal for built-in value and pointer types, and is equivalent to an atomic
     * "val = val - 1" operation.  This function exists to facilitate use of the optimized decrement
     * instructions provided by some architecures.  If no such instruction exists on the target platform
     * then the behavior will perform the operation using more traditional means.  This operation is both
     * lock-free and atomic.
     *
     * Returns:
     *  The result of an atomicLoad of val immediately following the increment operation.  This value
     *  is not required to be equal to the newly stored value.  Thus, competing writes are allowed to
     *  occur between the increment and successive load operation.
     *
     * Params:
     *  val = The value to decrement.
     */
    template atomicDecrement( T, msync ms : msync.none ) { alias doAtomicDecrement!(T,ms!=msync.none) atomicDecrement; }
    template atomicDecrement( T, msync ms : msync.ssb )  { alias doAtomicDecrement!(T,ms!=msync.none) atomicDecrement; } /// ditto
    template atomicDecrement( T, msync ms : msync.acq )  { alias doAtomicDecrement!(T,ms!=msync.none) atomicDecrement; } /// ditto
    template atomicDecrement( T, msync ms : msync.rel )  { alias doAtomicDecrement!(T,ms!=msync.none) atomicDecrement; } /// ditto
}


////////////////////////////////////////////////////////////////////////////////
// Atomic
////////////////////////////////////////////////////////////////////////////////


/**
 * This class represents a value which will be subject to competing access.  All accesses to
 * this value will be synchronized with main memory, and various memory barriers may be
 * employed for instruction ordering.  Any primitive type of size equal to or smaller than
 * the memory bus size is allowed, so 32-bit machines may use values with size <= int.sizeof
 * and 64-bit machines may use values with size <= long.sizeof.  The one exception to this
 * rule is that architectures that support DCAS will allow double-wide storeIf operations.
 * The 32-bit x86 architecture, for example, supports 64-bit storeIf operations.
 */
struct Atomic( T )
{
    static assert( isValidAtomicType!(T) );
/+
    template load( msync ms : msync.none )  { alias atomicLoad!(T,ms) load; }
    template load( msync ms : msync.hlb )   { alias atomicLoad!(T,ms) load; }
    template load( msync ms : msync.ddhlb ) { alias atomicLoad!(T,ms) load; }
    template load( msync ms : msync.acq )   { alias atomicLoad!(T,ms) load; }

    template store( msync ms : msync.none ) { alias atomicStore!(T,ms) store; }
    template store( msync ms : msync.ssb )  { alias atomicStore!(T,ms) store; }
    template store( msync ms : msync.acq )  { alias atomicStore!(T,ms) store; }
    template store( msync ms : msync.rel )  { alias atomicStore!(T,ms) store; }

    template storeIf( msync ms : msync.none ) { alias atomicStoreIf!(T,ms) store; }
    template storeIf( msync ms : msync.ssb )  { alias atomicStoreIf!(T,ms) store; }
    template storeIf( msync ms : msync.acq )  { alias atomicStoreIf!(T,ms) store; }
    template storeIf( msync ms : msync.rel )  { alias atomicStoreIf!(T,ms) store; }
+/
    /**
     * Refreshes the contents of this value from main memory.  This operation is both lock-free and atomic.
     *
     * Returns:
     *  The loaded value.
     */
    template load( msync ms : msync.none )  { T load() { return atomicLoad!(T,ms)( m_val ); } }
    template load( msync ms : msync.hlb )   { T load() { return atomicLoad!(T,ms)( m_val ); } } /// ditto
    template load( msync ms : msync.ddhlb ) { T load() { return atomicLoad!(T,ms)( m_val ); } } /// ditto
    template load( msync ms : msync.acq )   { T load() { return atomicLoad!(T,ms)( m_val ); } } /// ditto

    /**
     * Stores 'newval' to the memory referenced by this value.  This operation is both lock-free and atomic.
     *
     * Params:
     *  newval  = The value to store.
     */
    template store( msync ms : msync.none ) { void store( T newval ) { atomicStore!(T,ms)( m_val, newval ); } }
    template store( msync ms : msync.ssb )  { void store( T newval ) { atomicStore!(T,ms)( m_val, newval ); } } /// ditto
    template store( msync ms : msync.acq )  { void store( T newval ) { atomicStore!(T,ms)( m_val, newval ); } } /// ditto
    template store( msync ms : msync.rel )  { void store( T newval ) { atomicStore!(T,ms)( m_val, newval ); } } /// ditto

    /**
     * Stores 'newval' to the memory referenced by this value if val is equal to 'equalTo'.  This operation
     * is both lock-free and atomic.
     *
     * Returns:
     *  true if the store occurred, false if not.
     *
     * Params:
     *  newval  = The value to store.
     *  equalTo = The comparison value.
     */
    template storeIf( msync ms : msync.none ) { bool storeIf( T newval, T equalTo ) { return atomicStoreIf!(T,ms)( m_val, newval, equalTo ); } }
    template storeIf( msync ms : msync.ssb )  { bool storeIf( T newval, T equalTo ) { return atomicStoreIf!(T,ms)( m_val, newval, equalTo ); } } /// ditto
    template storeIf( msync ms : msync.acq )  { bool storeIf( T newval, T equalTo ) { return atomicStoreIf!(T,ms)( m_val, newval, equalTo ); } } /// ditto
    template storeIf( msync ms : msync.rel )  { bool storeIf( T newval, T equalTo ) { return atomicStoreIf!(T,ms)( m_val, newval, equalTo ); } } /// ditto

    version( DDoc ) {} else {
    static if( isValidNumericType!(T) )
    {
        /**
         * This operation is only legal for built-in value and pointer types, and is equivalent to an atomic
         * "val = val + 1" operation.  This function exists to facilitate use of the optimized increment
         * instructions provided by some architecures.  If no such instruction exists on the target platform
         * then the behavior will perform the operation using more traditional means.  This operation is both
         * lock-free and atomic.
         *
         * Returns:
         *  The result of an atomicLoad of val immediately following the increment operation.  This value
         *  is not required to be equal to the newly stored value.  Thus, competing writes are allowed to
         *  occur between the increment and successive load operation.
         */
        template increment( msync ms : msync.none ) { T increment() { return atomicIncrement!(T,ms)( m_val ); } }
        template increment( msync ms : msync.ssb )  { T increment() { return atomicIncrement!(T,ms)( m_val ); } } /// ditto
        template increment( msync ms : msync.acq )  { T increment() { return atomicIncrement!(T,ms)( m_val ); } } /// ditto
        template increment( msync ms : msync.rel )  { T increment() { return atomicIncrement!(T,ms)( m_val ); } } /// ditto

        /**
         * This operation is only legal for built-in value and pointer types, and is equivalent to an atomic
         * "val = val - 1" operation.  This function exists to facilitate use of the optimized decrement
         * instructions provided by some architecures.  If no such instruction exists on the target platform
         * then the behavior will perform the operation using more traditional means.  This operation is both
         * lock-free and atomic.
         *
         * Returns:
         *  The result of an atomicLoad of val immediately following the increment operation.  This value
         *  is not required to be equal to the newly stored value.  Thus, competing writes are allowed to
         *  occur between the increment and successive load operation.
         */
        template decrement( msync ms : msync.none ) { T decrement() { return atomicDecrement!(T,ms)( m_val ); } }
        template decrement( msync ms : msync.ssb )  { T decrement() { return atomicDecrement!(T,ms)( m_val ); } } /// ditto
        template decrement( msync ms : msync.acq )  { T decrement() { return atomicDecrement!(T,ms)( m_val ); } } /// ditto
        template decrement( msync ms : msync.rel )  { T decrement() { return atomicDecrement!(T,ms)( m_val ); } } /// ditto
    }}

private:
    T   m_val;
}


////////////////////////////////////////////////////////////////////////////////
// Support Code for Unit Tests
////////////////////////////////////////////////////////////////////////////////


private
{
    template testLoad( T, msync ms )
    {
        void testLoad( T val = T.init + 1 )
        {
            T          base;
            Atomic!(T) atom;

            assert( atom.load!(ms)() == base );
            base        = val;
            atom.m_val  = val;
            assert( atom.load!(ms)() == base );
        }
    }

    template testStore( T, msync ms )
    {
        void testStore( T val = T.init + 1 )
        {
            T          base;
            Atomic!(T) atom;

            assert( atom.m_val == base );
            base = val;
            atom.store!(ms)( base );
            assert( atom.m_val == base );
        }
    }

    template testStoreIf( T, msync ms )
    {
        void testStoreIf( T val = T.init + 1 )
        {
            T          base;
            Atomic!(T) atom;

            assert( atom.m_val == base );
            base = val;
            atom.storeIf!(ms)( base, val );
            assert( atom.m_val != base );
            atom.storeIf!(ms)( base, T.init );
            assert( atom.m_val == base );
        }
    }

    template testIncrement( T, msync ms )
    {
        void testIncrement( T val = T.init + 1 )
        {
            T          base = val;
            T          incr = val;
            Atomic!(T) atom;

            atom.m_val = val;
            assert( atom.m_val == base && incr == base );
            base = base + 1;
            incr = atom.increment!(ms)();
            assert( atom.m_val == base && incr == base );
        }
    }

    template testDecrement( T, msync ms )
    {
        void testDecrement( T val = T.init + 1 )
        {
            T          base = val;
            T          decr = val;
            Atomic!(T) atom;

            atom.m_val = val;
            assert( atom.m_val == base && decr == base );
            base = base - 1;
            decr = atom.decrement!(ms)();
            assert( atom.m_val == base && decr == base );
        }
    }

    template testType( T )
    {
        void testType( T val = T.init + 1 )
        {
            testLoad!(T, msync.none)( val );
            testLoad!(T, msync.acq)( val );

            testStore!(T, msync.none)( val );
            testStore!(T, msync.acq)( val );
            testStore!(T, msync.rel)( val );

            testStoreIf!(T, msync.none)( val );
            testStoreIf!(T, msync.acq)( val );

            static if( isValidNumericType!(T) )
            {
                testIncrement!(T, msync.none)( val );
                testIncrement!(T, msync.acq)( val );

                testDecrement!(T, msync.none)( val );
                testDecrement!(T, msync.acq)( val );
            }
        }
    }
}


////////////////////////////////////////////////////////////////////////////////
// Unit Tests
////////////////////////////////////////////////////////////////////////////////


unittest
{
    testType!(bool)();

    testType!(byte)();
    testType!(ubyte)();

    testType!(short)();
    testType!(ushort)();

    testType!(int)();
    testType!(uint)();

    int x;
    testType!(void*)( &x );

    //
    // long
    //

    version( X86_64 )
    {
        testLoad!(long, msync.none)();
        testLoad!(long, msync.acq)();

        testStore!(long, msync.none)();
        testStore!(long, msync.acq)();
        testStore!(long, msync.rel)();

        testIncrement!(long, msync.none)();
        testDecrement!(long, msync.acq)();
    }
    version( X86 )
    {
        testStoreIf!(long, msync.none)();
        testStoreIf!(long, msync.acq)();
    }

    //
    // ulong
    //

    version( X86_64 )
    {
        testLoad!(ulong, msync.none)();
        testLoad!(ulong, msync.acq)();

        testStore!(ulong, msync.none)();
        testStore!(ulong, msync.acq)();
        testStore!(ulong, msync.rel)();

        testIncrement!(ulong, msync.none)();
        testDecrement!(ulong, msync.acq)();
    }
    version( X86 )
    {
        testStoreIf!(ulong, msync.none)();
        testStoreIf!(ulong, msync.acq)();
    }
}